@@ -55,6 +55,25 @@ class BaseCloudNodeConfigurator:
 
     NAMESSPACE_CREATE_ACTIONS = ['add', 'create']
     application = 'ursula'
+    required_fields = [
+        'eth_provider',
+        'payment_provider',
+        'docker_image',
+        'payment_network'
+    ]
+
+    host_level_override_prompts = {
+        'eth_provider': {"prompt": "--eth-provider: please provide the url of a hosted ethereum node (infura/geth) which your nodes can access", "choices": None},
+        'payment_provider': {"prompt": "--payment-provider: please provide the url of a hosted level-two node (infura/bor) which your nodes can access", "choices": None},
+        'payment_network':  {"prompt": f'--payment-network:  choose a payment network from: {PAYMENT_NETWORK_CHOICES}', "choices": PAYMENT_NETWORKS},
+    }
+
+    output_capture = {
+        'operator address': [],
+        'rest url': [],
+        'nucypher version': [],
+        'nickname': []
+    }
 
     def __init__(self,  # TODO: Add type annotations
                  emitter,
@@ -99,11 +118,6 @@ class BaseCloudNodeConfigurator:
 
         self.created_new_nodes = False
 
-        # the keys in this dict are used as search patterns by the anisble result collector and it will return
-        # these values for each node if it happens upon them in some output
-        self.output_capture = {'operator address': [], 'rest url': [
-        ], 'nucypher version': [], 'nickname': []}
-
         if pre_config:
             self.config = pre_config
             self.namespace_network = self.config.get('namespace')
@@ -145,18 +159,12 @@ class BaseCloudNodeConfigurator:
         # if certain config options have been specified with this invocation,
         # save these to update host specific variables before deployment
         # to allow for individual host config differentiation
-        self.host_level_overrides = {
+        self.host_level_overrides = {k: v for k, v in {
             'eth_provider': eth_provider,
             'payment_provider': self.kwargs.get('payment_provider'),
             'docker_image': docker_image,
             'payment_network':  self.kwargs.get('payment_network'),
-        }
-
-        self.host_level_override_prompts = {
-            'eth_provider': {"prompt": "--eth-provider: please provide the url of a hosted ethereum node (infura/geth) which your nodes can access", "choices": None},
-            'payment_provider': {"prompt": "--payment-provider: please provide the url of a hosted level-two node (infura/bor) which your nodes can access", "choices": None},
-            'payment_network':  {"prompt": f'--payment-network:  choose a payment network from: {PAYMENT_NETWORK_CHOICES}', "choices": PAYMENT_NETWORKS},
-        }
+        }.items() if k in self.required_fields}
 
         self.config['seed_network'] = seed_network if seed_network is not None else self.config.get(
             'seed_network')
@@ -237,7 +245,7 @@ class BaseCloudNodeConfigurator:
                      self.config['ethpassword']),
                 ],
             'cliargs': [
-                    ]
+            ]
         }
 
         if generate_keymaterial or kwargs.get('migrate_nucypher') or kwargs.get('init'):
@@ -600,7 +608,7 @@ class BaseCloudNodeConfigurator:
                 read_only=True
             )
             yield (ns, dep.get_all_hosts())
-    
+
     def get_host_by_name(self, host_name):
         return next([host_data for node_name, host_data in self.get_all_hosts() if node_name == host_name])
 
@@ -610,8 +618,8 @@ class BaseCloudNodeConfigurator:
     def destroy_resources(self, node_names):
         node_names = [s for s in node_names if s in [
             names for names, data in self.get_provider_hosts()]]
-        if self.emitter.prompt( f"Destroying {self.provider_name} instances for nodes: {' '.join(node_names)}.  Continue?  (type 'yes')") == 'yes':
-            
+        if self.emitter.prompt(f"Destroying {self.provider_name} instances for nodes: {' '.join(node_names)}.  Continue?  (type 'yes')") == 'yes':
+
             if self._destroy_resources(node_names):
                 if not self.config.get('instances'):
                     self.emitter.echo(
@@ -657,8 +665,9 @@ class BaseCloudNodeConfigurator:
                     f"\t{node_name}: {host_data['publicaddress']}")
                 self.emitter.echo(
                     f"\t\t ssh: {dep.format_ssh_cmd(host_data)}", color="yellow")
-                self.emitter.echo(
-                    f"\t\t operator address: {host_data['operator address']}")
+                if host_data.get('operator address'):
+                    self.emitter.echo(
+                        f"\t\t operator address: {host_data['operator address']}")
 
         if backup:
             self.emitter.echo(
@@ -667,14 +676,14 @@ class BaseCloudNodeConfigurator:
                 f" Backup data can be found here: {self.config_dir}/remote_worker_backups/")
 
     def format_ssh_cmd(self, host_data):
-        
+
         keypair = ''
         user = next(v['value'] for v in host_data['provider_deploy_attrs']
                     if v['key'] == 'default_user')
 
         if any([pda['key'] == 'ansible_ssh_private_key_file' for pda in host_data['provider_deploy_attrs']]):
-            keypair ='-i "' + next(v['value'] for v in host_data['provider_deploy_attrs']
-                                if v['key'] == 'ansible_ssh_private_key_file') + '"'
+            keypair = '-i "' + next(v['value'] for v in host_data['provider_deploy_attrs']
+                                    if v['key'] == 'ansible_ssh_private_key_file') + '"'
 
         return f"ssh {user}@{host_data['publicaddress']} {keypair}"
 
@@ -693,7 +702,8 @@ class BaseCloudNodeConfigurator:
             if not self.config['instances'][instance].get('index'):
                 self.config['instances'][instance]['index'] = index
             if instance.runtime_envvars.get('NUCYPHER_WORKER_ETH_PASSWORD'):
-                instance.runtime_envvars['NUCYPHER_OPERATOR_ETH_PASSWORD'] = instance.runtime_envvars.get('NUCYPHER_WORKER_ETH_PASSWORD')
+                instance.runtime_envvars['NUCYPHER_OPERATOR_ETH_PASSWORD'] = instance.runtime_envvars.get(
+                    'NUCYPHER_WORKER_ETH_PASSWORD')
                 del instance.runtime_envvars['NUCYPHER_WORKER_ETH_PASSWORD']
 
         if self.config.get('keyringpassword'):
@@ -743,29 +753,44 @@ class DigitalOceanConfigurator(BaseCloudNodeConfigurator):
         ]
 
     def _configure_provider_params(self):
-        self.token = os.environ.get('DIGITALOCEAN_ACCESS_TOKEN')
+
+        self.token = self.config.get('digital-ocean-access-token')
         if not self.token:
             self.emitter.echo(
-                f"Please `export DIGITALOCEAN_ACCESS_TOKEN=<your access token.>` from here:  https://cloud.digitalocean.com/account/api/tokens\n", color="red")
-            raise AttributeError(
-                "Could not continue without DIGITALOCEAN_ACCESS_TOKEN environment variable.")
+                "checking envvar DIGITALOCEAN_ACCESS_TOKEN for access token...")
+            self.token = os.environ.get('DIGITALOCEAN_ACCESS_TOKEN')
+        if not self.token:
+            self.token = self.emitter.prompt(
+                f"Please enter your Digital Ocean Access Token which can be created here: https://cloud.digitalocean.com/account/api/tokens.  It looks like this: b34abcDEF17ABCDEFAbcDEf09fd72a28425ABCDEF8b198e9623ABCDEFc11591.  You can also `export DIGITALOCEAN_ACCESS_TOKEN=<the token>`")
+            if not self.token:
+                raise AttributeError(
+                    "Could not continue without token or DIGITALOCEAN_ACCESS_TOKEN environment variable.")
+
         self.region = os.environ.get('DIGITALOCEAN_REGION') or self.config.get(
             'digital-ocean-region') or self.kwargs.get('region') or self.default_region
 
         self.emitter.echo(
-            f'using DigitalOcean region: {self.region}, to change regions `export DIGITALOCEAN_REGION: https://www.digitalocean.com/docs/platform/availability-matrix/\n', color='yellow')
+            f'using DigitalOcean region: {self.region}, to change regions call this command with --region specified or `export DIGITALOCEAN_REGION: https://www.digitalocean.com/docs/platform/availability-matrix/\n', color='yellow')
 
-        self.sshkey = os.environ.get(
-            'DIGITAL_OCEAN_KEY_FINGERPRINT') or self.config.get('sshkey')
+        self.sshkey = self.config.get('sshkey')
         if not self.sshkey:
             self.emitter.echo(
-                "Please set the name of your Digital Ocean SSH Key (`export DIGITAL_OCEAN_KEY_FINGERPRINT=<your preferred ssh key fingerprint>` from here: https://cloud.digitalocean.com/account/security", color="red")
-            self.emitter.echo(
-                "it should look like `DIGITAL_OCEAN_KEY_FINGERPRINT=88:fb:53:51:09:aa:af:02:e2:99:95:2d:39:64:c1:64`\n", color="red")
-            raise AttributeError(
-                "Could not continue without DIGITAL_OCEAN_KEY_FINGERPRINT environment variable.")
+                "checking envvar DIGITAL_OCEAN_KEY_FINGERPRINT for key fingerprint...")
+            self.sshkey = os.environ.get('DIGITAL_OCEAN_KEY_FINGERPRINT')
+        if not self.sshkey:
+            self.sshkey = self.emitter.prompt(
+                f"Please `enter the fingerprint of a Digital Ocean SSH Key which can be created here: https://cloud.digitalocean.com/account/security.  They look like this: 8a:db:8f:4c:b1:61:fa:84:21:30:4d:d6:77:3b:a1:4d")
+            if not self.sshkey:
+                self.emitter.echo(
+                    "Please set the name of your Digital Ocean SSH Key (`export DIGITAL_OCEAN_KEY_FINGERPRINT=<your preferred ssh key fingerprint>` from here: https://cloud.digitalocean.com/account/security", color="red")
+                self.emitter.echo(
+                    "it should look like `DIGITAL_OCEAN_KEY_FINGERPRINT=88:fb:53:51:09:aa:af:02:e2:99:95:2d:39:64:c1:64`\n", color="red")
+                raise AttributeError(
+                    "Could not continue without DIGITAL_OCEAN_KEY_FINGERPRINT.")
+
         self.config['sshkey'] = self.sshkey
         self.config['digital-ocean-region'] = self.region
+        self.config['digital-ocean-access-token'] = self.token
 
         self._write_config()
 
@@ -821,7 +846,8 @@ class DigitalOceanConfigurator(BaseCloudNodeConfigurator):
 
     def _destroy_resources(self, node_names):
 
-        existing_instances = {k: v for k, v in self.get_provider_hosts() if k in node_names}
+        existing_instances = {
+            k: v for k, v in self.get_provider_hosts() if k in node_names}
         if existing_instances:
             self.emitter.echo(
                 f"\nAbout to destroy the following: {', '.join(existing_instances.keys())}, including all local data about these nodes.")
@@ -1258,9 +1284,124 @@ class GenericConfigurator(BaseCloudNodeConfigurator):
 
         return self.config
 
-    def deploy_image_on_existing_nodes(self, hosts, resource):
-        self.emitter.echo("this feature is a work in progress")
-        pass
+
+class PorterDeployer(BaseCloudNodeConfigurator):
+
+    application = 'porter'
+    required_fields = [
+        'eth_provider',
+        'docker_image',
+    ]
+    host_level_override_prompts = {
+        'eth_provider': {"prompt": "--eth-provider: please provide the url of a hosted ethereum node (infura/geth) which this porter node can access", "choices": None},
+    }
+
+    output_capture = {}
+
+    @property
+    def _inventory_template(self):
+        template_path = Path(TEMPLATES).joinpath('porter_inventory.mako')
+        return Template(filename=str(template_path))
+
+    @property
+    def inventory_path(self):
+        return str(Path(DEFAULT_CONFIG_ROOT).joinpath(NODE_CONFIG_STORAGE_KEY, f'{self.namespace_network}.porter_ansible_inventory.yml'))
+
+    def deploy_porter_on_existing_nodes(self, node_names):
+        playbook = Path(PLAYBOOKS).joinpath('setup_porter.yml')
+        self.configure_host_level_overrides(node_names)
+
+        self.update_generate_inventory(node_names)
+
+        for k in node_names:
+            installed = self.config['instances'][k].get('installed', [])
+            installed = list(set(installed + [self.application]))
+            self.config['instances'][k]['installed'] = installed
+        self._write_config()
+
+        loader = DataLoader()
+        inventory = InventoryManager(
+            loader=loader, sources=self.inventory_path)
+        callback = AnsiblePlayBookResultsCollector(
+            sock=self.emitter, return_results=self.output_capture)
+        variable_manager = VariableManager(loader=loader, inventory=inventory)
+
+        executor = PlaybookExecutor(
+            playbooks=[playbook],
+            inventory=inventory,
+            variable_manager=variable_manager,
+            loader=loader,
+            passwords=dict(),
+        )
+        executor._tqm._stdout_callback = callback
+        executor.run()
+
+        for k in node_names:
+            installed = self.config['instances'][k].get('installed', [])
+            installed = list(set(installed + [self.application]))
+            self.config['instances'][k]['installed'] = installed
+        self._write_config()
+
+        self.update_captured_instance_data(self.output_capture)
+        self.give_helpful_hints(node_names, backup=True, playbook=playbook)
+
+    def update_generate_inventory(self, node_names, **kwargs):
+
+        # filter out the nodes we will not be dealing with
+        nodes = {key: value for key,
+                 value in self.config['instances'].items() if key in node_names}
+        if not nodes:
+            raise KeyError(
+                f"No hosts matched the supplied names: {node_names}.  Try `nucypher-ops nodes list` or create new hosts with `nucypher-ops nodes create`")
+
+        defaults = {
+            'envvars':
+                [
+
+                ],
+            'cliargs': [
+            ]
+        }
+        for datatype in ['envvars', 'cliargs']:
+
+            data_key = f'runtime_{datatype}'
+
+            input_data = [(k, v) for k, v in getattr(self, datatype)]
+
+            # populate the specified environment variables as well as the
+            # defaults that are only used in the inventory
+            for key, node in nodes.items():
+                node_vars = nodes[key].get(data_key, {})
+                for k, v in input_data:
+                    node_vars.update({k: v})
+                nodes[key][data_key] = node_vars
+
+                # we want to update the config with the specified values
+                # so they will persist in future invocations
+                self.config['instances'][key] = copy.deepcopy(nodes[key])
+
+            # we don't want to save the default_envvars to the config file
+            # but we do want them to be specified to the inventory template
+            # but overridden on a per node basis if previously specified
+            for key, node in nodes.items():
+                for k, v in defaults[datatype]:
+                    if not k in nodes[key][data_key]:
+                        nodes[key][data_key][k] = v
+
+        inventory_content = self._inventory_template.render(
+            deployer=self,
+            nodes=nodes.values(),
+            extra=kwargs,
+        )
+
+        with open(self.inventory_path, 'w') as outfile:
+            outfile.write(inventory_content)
+            self.emitter.echo(f"wrote new inventory to: {self.inventory_path}")
+
+        # now that everything rendered correctly, save how we got there.
+        self._write_config()
+
+        return self.inventory_path
 
 
 class CloudDeployers:
@@ -1268,6 +1409,7 @@ class CloudDeployers:
     aws = AWSNodeConfigurator
     digitalocean = DigitalOceanConfigurator
     generic = GenericConfigurator
+    porter = PorterDeployer
 
     @staticmethod
     def get_deployer(name):
